import logging, json, os, nre, sequtils, strtabs, tables
import private/util

# Types
#
type
  BuildStatus* = object
    state*, details*: string

  Step* = object
    name*, stepCmd*, workingDir*: string
    artifacts*, cmdInput*, depends*, expectedEnv*: seq[string]
    dontSkip*: bool

  ProjectCfg* = object
    name*: string
    versionCmd*: string
    steps*: Table[string, Step]

  ProjectDef* = object
    cfgFilePath*, defaultBranch*, name*, repo*: string
    envVars*: StringTableRef

  StrawBossConfig* = object
    artifactsRepo*: string
    projects*: seq[ProjectDef]

  RunRequest* = object
    projectName*, stepName*, buildRef*, workspaceDir*: string
    forceRebuild*: bool

# internal utils

let nullNode = newJNull()
proc getIfExists(n: JsonNode, key: string): JsonNode =
  result = if n.hasKey(key): n[key]
           else: nullNode

proc getOrFail(n: JsonNode, key: string, objName: string = ""): JsonNode =
  if not n.hasKey(key): raiseEx objName & " missing key " & key
  return n[key]

# Configuration parsing code

proc loadStrawBossConfig*(cfgFile: string): StrawBossConfig =
  if not existsFile(cfgFile):
    raiseEx "strawboss config file not found: " & cfgFile

  let jsonCfg = parseFile(cfgFile)

  var projectDefs: seq[ProjectDef] = @[]

  for pJson in jsonCfg.getIfExists("projects").getElems:
    var envVars = newStringTable(modeCaseSensitive)
    for k, v in pJson.getIfExists("envVars").getFields: envVars[k] = v.getStr("")

    projectDefs.add(
      ProjectDef(
        cfgFilePath: pJson.getIfExists("cfgFilePath").getStr("strawboss.json"),
        defaultBranch: pJson.getIfExists("defaultBranch").getStr("master"),
        name: pJson.getOrFail("name", "project definition").getStr,
        envVars: envVars,
        repo: pJson.getOrFail("repo", "project definition").getStr))

  result = StrawBossConfig(
    artifactsRepo: jsonCfg.getIfExists("artifactsRepo").getStr("artifacts"),
    projects: projectDefs)

proc loadProjectConfig*(cfgFile: string): ProjectCfg =
  if not existsFile(cfgFile):
    raiseEx "project config file not found: " & cfgFile

  let jsonCfg = parseFile(cfgFile)

  if not jsonCfg.hasKey("steps"):
    raiseEx "project configuration is missing steps definition"

  var steps = initTable[string, Step]()
  for sName, pJson in jsonCfg.getOrFail("steps", "project configuration").getFields:
    steps[sName] = Step(
      name:         sName,
      workingDir:   pJson.getIfExists("workingDir").getStr("."),
      stepCmd:      pJson.getIfExists("stepCmd").getStr("sh"),
      depends:      pJson.getIfExists("depends").getElems.mapIt(it.getStr),
      artifacts:    pJson.getIfExists("artifacts").getElems.mapIt(it.getStr),
      cmdInput:     pJson.getIfExists("cmdInput").getElems.mapIt(it.getStr),
      expectedEnv:  pJson.getIfExists("expectedEnv").getElems.mapIt(it.getStr),
      dontSkip:     pJson.getIfExists("dontSkip").getStr("false") != "false")

    if steps[sName].stepCmd == "sh" and steps[sName].cmdInput.len == 0:
      warn "Step " & sName & " uses 'sh' as its command but has no cmdInput."

  result = ProjectCfg(
    name: jsonCfg.getOrFail("name", "project configuration").getStr,
    versionCmd: jsonCfg.getIfExists("versionCmd").getStr("git describe --tags --always"),
    steps: steps)

proc loadBuildStatus*(statusFile: string): BuildStatus =
  if not existsFile(statusFile): raiseEx "status file not found: " & statusFile
  let jsonObj = parseFile(statusFile)

  result = BuildStatus(
    state: jsonObj.getOrFail("state", "build status").getStr,
    details: jsonObj.getIfExists("details").getStr("") )

proc parseRunRequest*(reqStr: string): RunRequest =
  let reqJson = parseJson(reqStr)

  result = RunRequest(
    projectName: reqJson.getOrFail("projectName", "RunRequest").getStr,
    stepName: reqJson.getOrFail("stepName", "RunRequest").getStr,
    buildRef: reqJson.getOrFail("buildRef", "RunRequest").getStr,
    workspaceDir: reqJson.getOrFail("workspaceDir", "RunRequest").getStr,
    forceRebuild: reqJson.getOrFail("forceRebuild", "RunRequest").getBVal)

# TODO: can we use the marshal module for this?
proc `%`*(s: BuildStatus): JsonNode =
  result = %* {
    "state": s.state,
    "details": s.details
  }

proc `%`*(req: RunRequest): JsonNode =
  result = %* {
    "projectName": req.projectName,
    "stepName": req.stepName,
    "buildRef": req.buildRef,
    "workspaceDir": req.workspaceDir,
    "forceRebuild": req.forceRebuild
    }

proc `$`*(s: BuildStatus): string = result = pretty(%s)
proc `$`*(req: RunRequest): string = result = pretty(%req)
